==> rasa-x-0.42.0.dev20210708.tar.gz: rasa-x-0.42.0.dev20210708/rasax/community/services/workspaces_service.py <==
from typing import Any, Dict, List, Optional, Text

import rasax.community.config as rasa_x_config
from rasax.community.database.service import DbService
from rasax.community.database.workspace import Workspace
from rasax.community.services.logs_service import LogsService
from sanic.request import Request


class WorkspacesError(Exception):
    """Base exception for all Workspaces errors."""


class WorkspacesNameExists(WorkspacesError):
    """Exception that is raised when there's already a workspace with a given name."""


class WorkspacesIdNotExists(WorkspacesError):
    """Exception that is raised when there's no workspace with a given ID."""


class WorkspaceIsDefaultWorkspace(WorkspacesError):
    """Exception that is raised when user tries to rename or delete a default workspace."""


class WorkspacesDraftsCannotReject(WorkspacesError):
    """Exception that is raised when drafts cannot be rejected."""


class WorkspacesDraftsCannotAccept(WorkspacesError):
    """Exception that is raised when drafts cannot be accepted."""


class WorkspacesService(DbService):
    """Service for all Workspaces related SQL operations."""

    @staticmethod
    def from_request(request: Request, **kwargs) -> "WorkspacesService":
        """Creates a `WorkspacesService` from an incoming HTTP request's DB session.

        Args:
            request: Incoming HTTP request.
            **kwargs: other key-value args, not used.

        Returns:
            `WorkspacesService` instance with a connection to the DB.
        """
        return WorkspacesService(request.ctx.db_session)

    def create_workspace(self, name: Text) -> Dict[Text, Any]:
        """Creates a workspace with a given name.

        Args:
            name: Name to assign to the new workspace.

        Raises:
            ValueError: If the workspace name is already taken.

        Returns:
            New workspace.
        """
        workspace = self.query(Workspace).filter(Workspace.name == name).first()
        if workspace:
            raise WorkspacesNameExists(f"Workspace with name '{name}' already exists.")

        workspace = Workspace(name=name)
        self.add(workspace)
        self.flush()

        return workspace.as_dict()

    def get_workspaces(self) -> List[Dict[Text, Any]]:
        """Retrieves a list of existing workspaces.

        Returns:
           Existing workspaces.
        """
        return [workspace.as_dict() for workspace in self.query(Workspace)]

    def update_workspace(self, workspace_id: int, name: Text) -> Dict[Text, Any]:
        """Updates workspace name by id.

        Args:
            workspace_id: ID of the workspace to be updated.
            name: New name to assign to the workspace.

        Raises:
            WorkspacesNameExists: If a workspace with the same name already exists.
            WorkspacesIdNotExists: If a workspace with the ID doesn't exists.
            WorkspaceIsDefaultWorkspace: If a workspace is a default workspace and cannot be renamed.

        Returns:
            Updated workspace.
        """
        # check if there is already another workspace with the same name
        workspace_same_name = (
            self.query(Workspace).filter(Workspace.name == name).first()
        )
        if workspace_same_name:
            raise WorkspacesNameExists(f"Workspace with name '{name}' already exists.")

        # check that workspace to be renamed exists
        workspace = self.query(Workspace).filter(Workspace.id == workspace_id).first()
        if not workspace:
            raise WorkspacesIdNotExists(f"Workspace with id {workspace_id} not found.")

        # check if this is a user's default workspace
        if self.is_a_default_workspace(workspace_id):
            raise WorkspaceIsDefaultWorkspace(
                f"Workspace with id {workspace_id} is a default workspace."
            )

        workspace.name = name
        return workspace.as_dict()

    def delete_workspace(
        self, workspace_id: int, delete_default_workspace: bool = False
    ):
        """Deletes a workspace.

        Args:
            workspace_id: ID of the workspace to be deleted.
            delete_default_workspace: `True` if default workspace can be deleted.

        Raises:
            WorkspacesIdNotExists: If a workspace with the ID doesn't exists.
            WorkspaceIsDefaultWorkspace: If a workspace is a default workspace
                and cannot be renamed.

        Returns:
            Deleted workspace.
        """
        workspace = self.query(Workspace).filter(Workspace.id == workspace_id).first()
        if not workspace:
            raise WorkspacesIdNotExists(f"Workspace with id {workspace_id} not found.")

        if not delete_default_workspace and self.is_a_default_workspace(workspace_id):
            raise WorkspaceIsDefaultWorkspace(
                f"Workspace with id {workspace_id} is a default workspace."
            )

        # make sure drafts are moved back to their original state
        logs_service = LogsService(self.session)
        workspace_drafts = logs_service.fetch_logs(workspace_id=workspace_id)
        draft_ids = [d["id"] for d in workspace_drafts.result]
        self.reject_drafts_from_workspace(draft_ids)

        deleted_workspace = workspace.as_dict()
        self.delete(workspace)

        return deleted_workspace

    def is_a_default_workspace(self, workspace_id: int) -> bool:
        """Checks if workspace is any user's default workspace.

        Args:
            workspace_id: ID of workspace that potentially is a default workspace.

        Returns:
            `True` in case the workspace is a default workspace for some user.
        """
        from rasax.community.services.user_service import UserService

        user_service = UserService(self.session)
        users = user_service.fetch_all_users(rasa_x_config.team_name)
        for user in users:
            if user["default_workspace_id"] == int(workspace_id):
                return True
        return False

    def get_workspace_by_id(self, id: int) -> Optional[Dict[Text, Any]]:
        """Retrieves existing workspace by ID.

        Args:
            id: ID of the workspace.

        Returns:
            List of existing workspaces.
        """
        workspace = self.query(Workspace).filter(Workspace.id == id).first()
        return workspace.as_dict() if workspace else None

    def reject_drafts_from_workspace(self, draft_ids: List[int]) -> List[int]:
        """Rejects drafts from a workspace.

        Args:
            draft_ids: IDs of the drafts to be rejected.

        Returns:
            IDs of rejected drafts.
        """
        from rasax.community.services.logs_service import LogsService

        result = []
        logs_service = LogsService(self.session)
        try:
            for draft_id in draft_ids:
                result.append(
                    logs_service.restore_original_log_state(draft_id).get("id")
                )
        except ValueError as e:
            raise WorkspacesDraftsCannotReject(f"Failed to reject drafts, error: {e}")

        return result

    def accept_drafts_in_workspace(
        self, draft_ids: List[int], user: Dict[Text, Any]
    ) -> List[int]:
        """Accepts drafts in a workspace.

        This method will create new training data examples from the specified drafts,
        and "archive" the drafts themselves.

        Args:
            draft_ids: IDs of the drafts to be accepted.
            user: User who performs the action.

        Returns:
            IDs of accepted drafts.
        """
        from rasax.community.services.data_service import DataService
        from rasax.community.services.logs_service import LogsService

        project_name = rasa_x_config.project_name

        logs_service = LogsService(self.session)
        data_service = DataService(self.session)
        result = []
        for draft_id in draft_ids:
            log = logs_service.get_log_by_id(draft_id)
            if not log:
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: draft was not found."
                )
            if not log.workspace_id:
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: "
                    "draft doesn't have assigned workspace."
                )
            log_json = log.as_dict().get("user_input")
            if (
                not log_json
                or not log_json.get("intent")
                or not log_json.get("intent").get("name")
            ):
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: "
                    "draft doesn't have user parse data."
                )
            log_json["intent"] = log_json.get("intent").get("name")
            created_example = data_service.add_single_example(
                log_json, project_name, user
            )
            if not created_example:
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: "
                    "cannot create training example."
                )
            # Archive this draft, and remove it from the workspace
            log.in_training_data = True
            log.archived = True
            log.workspace_id = None
            result.append(log.id)
        return result

==> rasa-x-0.43.0.dev20211020.tar.gz: rasa-x-0.43.0.dev20211020/rasax/community/services/workspaces_service.py <==
import time
from typing import Any, Dict, List, Optional, Text

import rasax.community.constants
import rasax.community.config as rasa_x_config
from rasax.community.database.conversation import MessageLog
from rasax.community.database.service import DbService
from rasax.community.database.workspace import Workspace
from rasax.community.services.logs_service import LogsService
from sanic.request import Request


class WorkspacesError(Exception):
    """Base exception for all Workspaces errors."""


class WorkspacesNameExists(WorkspacesError):
    """Exception that is raised when there's already a workspace with a given name."""


class WorkspacesIdNotExists(WorkspacesError):
    """Exception that is raised when there's no workspace with a given ID."""


class WorkspaceIsDefaultWorkspace(WorkspacesError):
    """Exception that is raised when user tries to rename/delete a default workspace."""


class WorkspacesDraftsCannotReject(WorkspacesError):
    """Exception that is raised when drafts cannot be rejected."""


class WorkspacesDraftsCannotAccept(WorkspacesError):
    """Exception that is raised when drafts cannot be accepted."""


class WorkspacesService(DbService):
    """Service for all Workspaces related SQL operations."""

    @staticmethod
    def from_request(request: Request, **kwargs) -> "WorkspacesService":
        """Creates a `WorkspacesService` from an incoming HTTP request's DB session.

        Args:
            request: Incoming HTTP request.
            **kwargs: other key-value args, not used.

        Returns:
            `WorkspacesService` instance with a connection to the DB.
        """
        return WorkspacesService(request.ctx.db_session)

    def create_workspace(self, name: Text, username: Text) -> Dict[Text, Any]:
        """Creates a workspace with a given name.

        Args:
            name: Name to assign to the new workspace.
            username: User creating the new workspace.

        Raises:
            WorkspacesNameExists: If the workspace name is already taken.

        Returns:
            New workspace.
        """
        workspace = self.query(Workspace).filter(Workspace.name == name).first()
        if workspace:
            raise WorkspacesNameExists(f"Workspace with name '{name}' already exists.")

        workspace = Workspace(name=name)
        self.add(workspace)

        workspace.created_at = time.time()
        workspace.last_modifier_id = username

        # make sure we have `id`
        self.session.flush()

        return workspace.as_dict()

    def get_workspaces(self) -> List[Dict[Text, Any]]:
        """Retrieves a list of existing workspaces.

        Returns:
           Existing workspaces.
        """
        return [workspace.as_dict() for workspace in self.query(Workspace)]

    def update_workspace_name(
        self, workspace_id: int, name: Text, username: Text
    ) -> Dict[Text, Any]:
        """Updates workspace name by id.

        Args:
            workspace_id: ID of the workspace to be updated.
            name: New name to assign to the workspace.
            username: User updating the workspace name.

        Raises:
            WorkspacesNameExists: If a workspace with the same name already exists.
            WorkspacesIdNotExists: If a workspace with the ID doesn't exist.
            WorkspaceIsDefaultWorkspace: If a workspace is a default workspace
                and cannot be renamed.

        Returns:
            Updated workspace.
        """
        # check if there is already another workspace with the same name
        workspace_same_name = (
            self.query(Workspace).filter(Workspace.name == name).first()
        )
        if workspace_same_name:
            raise WorkspacesNameExists(f"Workspace with name '{name}' already exists.")

        # check that workspace to be renamed exists
        workspace = self.query(Workspace).filter(Workspace.id == workspace_id).first()
        if not workspace:
            raise WorkspacesIdNotExists(f"Workspace with id {workspace_id} not found.")

        # check if this is a user's default workspace
        if self.is_a_default_workspace(workspace_id):
            raise WorkspaceIsDefaultWorkspace(
                f"Workspace with id {workspace_id} is a default workspace."
            )

        workspace.name = name
        self.update_workspace_last_modified(workspace_id, username)

        return workspace.as_dict()

    def update_workspace_last_modified(
        self, workspace_id: int, username: Text
    ) -> Dict[Text, Any]:
        """Updates workspace `last_modified_at` and `last_modifier_id` by id.

        Args:
            workspace_id: ID of the workspace to be updated.
            username: User modifying the workspace.

        Raises:
            WorkspacesIdNotExists: If a workspace with the ID doesn't exist.

        Returns:
            Updated workspace.
        """
        workspace = self.query(Workspace).filter(Workspace.id == workspace_id).first()
        if not workspace:
            raise WorkspacesIdNotExists(f"Workspace with id {workspace_id} not found.")

        workspace.last_modified_at = time.time()
        workspace.last_modifier_id = username

        return workspace.as_dict()

    def delete_workspace(
        self, workspace_id: int, delete_default_workspace: bool = False
    ):
        """Deletes a workspace.

        Args:
            workspace_id: ID of the workspace to be deleted.
            delete_default_workspace: `True` if default workspace can be deleted.

        Raises:
            WorkspacesIdNotExists: If a workspace with the ID doesn't exist.
            WorkspaceIsDefaultWorkspace: If a workspace is a default workspace
                and cannot be deleted.

        Returns:
            Deleted workspace.
        """
        workspace = self.query(Workspace).filter(Workspace.id == workspace_id).first()
        if not workspace:
            raise WorkspacesIdNotExists(f"Workspace with id {workspace_id} not found.")

        if not delete_default_workspace and self.is_a_default_workspace(workspace_id):
            raise WorkspaceIsDefaultWorkspace(
                f"Workspace with id {workspace_id} is a default workspace."
            )

        # make sure drafts are moved back to their original state
        logs_service = LogsService(self.session)
        workspace_drafts = logs_service.fetch_logs(workspace_id=workspace_id)
        draft_ids = [d["id"] for d in workspace_drafts.result]
        self.reject_drafts_from_workspace(draft_ids)

        deleted_workspace = workspace.as_dict()
        self.delete(workspace)

        return deleted_workspace

    def is_a_default_workspace(self, workspace_id: int) -> bool:
        """Checks if workspace is any user's default workspace.

        Args:
            workspace_id: ID of workspace that potentially is a default workspace.

        Returns:
            `True` in case the workspace is a default workspace for some user.
        """
        from rasax.community.services.user_service import UserService

        user_service = UserService(self.session)
        users = user_service.fetch_all_users(rasa_x_config.team_name)
        for user in users:
            if user["default_workspace_id"] == int(workspace_id):
                return True
        return False

    def get_workspace_by_id(self, id: int) -> Optional[Dict[Text, Any]]:
        """Retrieves existing workspace by ID.

        Args:
            id: ID of the workspace.

        Returns:
            List of existing workspaces.
        """
        workspace = self.query(Workspace).filter(Workspace.id == id).first()
        if not workspace:
            raise WorkspacesIdNotExists(f"Workspace with id {id} not found.")

        return workspace.as_dict()

    def reject_drafts_from_workspace(
        self, draft_ids: List[int], username: Optional[Text] = None
    ) -> List[int]:
        """Rejects drafts from a workspace.

        Args:
            draft_ids: IDs of the drafts to be rejected.
            username: User rejecting drafts from the workspace.

        Returns:
            IDs of rejected drafts.
        """
        from rasax.community.services.logs_service import LogsService

        result = []
        logs_service = LogsService(self.session)
        workspace_ids = set()

        try:
            for draft_id in draft_ids:
                draft = logs_service.get_log_by_id(draft_id)
                if draft:
                    workspace_ids.add(draft.workspace_id)

                result.append(
                    logs_service.restore_original_log_state(draft_id).get("id")
                )
        except ValueError as e:
            raise WorkspacesDraftsCannotReject(f"Failed to reject drafts, error: {e}")

        if username:
            for workspace_id in workspace_ids:
                self.update_workspace_last_modified(workspace_id, username)

        return result

    def accept_all_drafts_in_workspace(self, workspace_id: int, user: Dict[Text, Any]):
        """Accepts all drafts in a workspace.

        This method will create new training data examples from each of the drafts
        belonging to the specified `workspace_id`.


        Args:
            draft_ids: IDs of the drafts to be accepted.
            user: User who performs the action.

        Returns:
            IDs of accepted drafts.
        """
        if not self.get_workspace_by_id(workspace_id):
            raise WorkspacesIdNotExists(
                f"Failed to accept drafts in the workspace. "
                f"Workspace with id {workspace_id} doesn't exist."
            )
        drafts = (
            self.query(MessageLog).filter(MessageLog.workspace_id == workspace_id).all()
        )
        draft_ids = [draft.id for draft in drafts]
        return self.accept_drafts_in_workspace(draft_ids, user)

    def accept_drafts_in_workspace(
        self, draft_ids: List[int], user: Dict[Text, Any]
    ) -> List[int]:
        """Accepts drafts in a workspace.

        This method will create new training data examples from the specified drafts,
        and "archive" the drafts themselves.

        Args:
            draft_ids: IDs of the drafts to be accepted.
            user: User who performs the action.

        Returns:
            IDs of accepted drafts.
        """
        from rasax.community.services.data_service import (
            DataService,
            TrainingDataCreationFailed,
        )
        from rasax.community.services.logs_service import LogsService

        project_name = rasa_x_config.project_name

        logs_service = LogsService(self.session)
        data_service = DataService(self.session)
        result = []
        workspace_ids = set()

        for draft_id in draft_ids:
            log = logs_service.get_log_by_id(draft_id)
            if not log:
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: draft was not found."
                )
            if not log.workspace_id:
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: "
                    "draft doesn't have assigned workspace."
                )
            workspace_ids.add(log.workspace_id)
            log_json = log.as_dict().get("user_input")
            if (
                not log_json
                or not log_json.get("intent")
                or not log_json.get("intent").get("name")
            ):
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: "
                    "draft doesn't have user parse data."
                )
            log_json["intent"] = log_json.get("intent").get("name")
            try:
                data_service.add_single_example(log_json, project_name, user)
            except TrainingDataCreationFailed as e:
                raise WorkspacesDraftsCannotAccept(
                    f"Failed to accept draft with id {draft_id}: "
                    "Cannot create training example. "
                    f"Error: {e}",
                )
            # Archive this draft, and remove it from the workspace
            log.in_training_data = True
            log.archived = True
            log.workspace_id = None
            result.append(log.id)

        for workspace_id in workspace_ids:
            self.update_workspace_last_modified(
                workspace_id, user[rasax.community.constants.USERNAME_KEY]
            )

        return result
